/*
 *
 * Copyright 2017 gRPC authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// Package roundrobin defines a roundrobin balancer. Roundrobin balancer is
// installed as one of the default balancers in gRPC, users don't need to
// explicitly install this balancer.
package random

import (
	"google.golang.org/grpc/balancer"
	"google.golang.org/grpc/balancer/base"
	"google.golang.org/grpc/grpclog"
	"log"
	"math/rand"
)

// Name is the name of round_robin balancer.
const Name = "my_random_robin"

var logger = grpclog.Component("myrandomrobin")

// newBuilder creates a new roundrobin balancer builder.
func newBuilder() balancer.Builder {
	return base.NewBalancerBuilder(Name, &randomPickerBuilder{}, base.Config{HealthCheck: true})
}

func init() {
	balancer.Register(newBuilder())
}

type randomPickerBuilder struct{}

type subConnInfo struct {
	conn     balancer.SubConn
	connInfo base.SubConnInfo
}

func (r *randomPickerBuilder) Build(info base.PickerBuildInfo) balancer.Picker {
	logger.Infof("myrandomrobin Picker: Build called with info: %v", info)
	log.Printf("myrandomrobin Picker: Build called with info: %v", info)
	if len(info.ReadySCs) == 0 {
		return base.NewErrPicker(balancer.ErrNoSubConnAvailable)
	}
	scs := make([]*subConnInfo, 0, len(info.ReadySCs))

	for sc, scInfo := range info.ReadySCs {
		scs = append(scs, &subConnInfo{
			conn:     sc,
			connInfo: scInfo,
		})
	}
	return &randomPicker{
		subConns: scs,
	}
}

type randomPicker struct {
	subConns []*subConnInfo
}

func (p *randomPicker) Pick(balancer.PickInfo) (balancer.PickResult, error) {
	sc := p.subConns[rand.Intn(len(p.subConns))]
	log.Printf("randomPicker Pick: SubConn: %s", sc.connInfo.Address.String())
	return balancer.PickResult{SubConn: sc.conn}, nil
}
